"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createScope = exports.useScope = void 0;
const create_js_1 = require("./create.js");
const frame_js_1 = require("./frame.js");
const instructions_js_1 = require("../instructions.js");
/**
 * Get the scope of the currently running {@link Operation}.
 *
 * @returns an operation yielding the current scope
 */
function* useScope() {
    let frame = yield* (0, instructions_js_1.getframe)();
    let [scope] = createScope(frame);
    return scope;
}
exports.useScope = useScope;
/**
 * Create a new scope to serve as an entry point from normal
 * JavaScript execution into Effection.
 *
 * When creating a fresh scope (as opposed to capturing a reference to
 * an existing one via {@link useScope}) it is the responsibility of
 * the creator of the new scope to destroy it when it is no longer needed.
 *
 * @example
 * ```js
 * let [scope, destroy] = createScope();
 * let task = scope.run(function*() {
 *   //do some long running work
 * });
 *
 * //... later
 * await destroy();
 *
 * ```
 * @returns a tuple containing the new scope, and a function to destroy it.
 */
function createScope(frame) {
    let parent = frame ?? (0, frame_js_1.createFrame)({ operation: instructions_js_1.suspend });
    let scope = (0, create_js_1.create)("Scope", {}, {
        run(operation) {
            let frame = parent.createChild(operation);
            frame.enter();
            return frame.getTask();
        },
        get(context) {
            let { key, defaultValue } = context;
            return (parent.context[key] ?? defaultValue);
        },
        set(context, value) {
            let { key } = context;
            parent.context[key] = value;
            return value;
        },
    });
    parent.enter();
    return [scope, parent.getTask().halt];
}
exports.createScope = createScope;
//# sourceMappingURL=scope.js.map