"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSignal = exports.SignalQueueFactory = void 0;
const queue_js_1 = require("./queue.js");
const instructions_js_1 = require("./instructions.js");
const context_js_1 = require("./context.js");
/**
 * @ignore
 * {@link Context} that contains a {@link Queue} factory to be used when creating a {@link Signal}.
 *
 * This allows end-users to customize a Signal's Queue.
 *
 * @example
 * ```javascript
 * export function useActions(pattern: ActionPattern): Stream<AnyAction, void> {
 *  return {
 *    *subscribe() {
 *      const actions = yield* ActionContext;
 *      yield* QueueFactory.set(() => createFilterQueue(matcher(pattern));
 *      return yield* actions.subscribe();
 *    }
 *  }
 * }
 *
 * function createFilterQueue(predicate: Predicate) {
 *  let queue = createQueue();
 *
 *  return {
 *    ...queue,
 *    add(value) {
 *      if (predicate(value)) {
 *        queue.add(value);
 *      }
 *    }
 *  }
 * }
 * ```
 */
exports.SignalQueueFactory = (0, context_js_1.createContext)("Signal.createQueue", queue_js_1.createQueue);
/**
 * Create a new {@link Signal}
 *
 * Signal should be used when you need to send messages to a stream
 * from _outside_ of an operation. The most common case of this is to
 * connect a plain, synchronous JavaScript callback to an operation.
 *
 * @example
 * ```javascript
 * function* logClicks(button) {
 *   let clicks = createSignal<MouseEvent>();
 *   try {
 *     button.addEventListener("click", clicks.send);
 *
 *     for (let click of yield* each(clicks)) {
 *       console.log("click", click);
 *     }
 *    } finally {
 *      button.removeEventListener("click", clicks.send);
 *    }
 * }
 * ```
 *
 * Do not use a signal to send messages from within an operation as it could
 * result in out-of-scope code being executed. In those cases, you should use a
 * {@link Channel}.
 */
function createSignal() {
    let subscribers = new Set();
    let useSubscription = (0, instructions_js_1.resource)(function* (provide) {
        let newQueue = yield* exports.SignalQueueFactory;
        let queue = newQueue();
        subscribers.add(queue);
        try {
            yield* provide({ next: queue.next });
        }
        finally {
            subscribers.delete(queue);
        }
    });
    function send(value) {
        for (let queue of [...subscribers]) {
            queue.add(value);
        }
    }
    function close(value) {
        for (let queue of [...subscribers]) {
            queue.close(value);
        }
    }
    return { send, close, subscribe: () => useSubscription };
}
exports.createSignal = createSignal;
//# sourceMappingURL=signal.js.map